# File: _names.py
# Created Date: Wednesday June 22nd 2022
# Author: Steven Atkinson (steven@atkinson.mn)


"""
Layer names
"""

ACTIVATION_NAME = "activation"
BATCHNORM_NAME = "batchnorm"
CONV_NAME = "conv"
LAYERNORM_NAME = "layernorm"
